# Development Workflow

Contributing is encouraged if you have calibration notebooks that you would like to add to the pipeline
or if you have a specific algorithm which is beneficial for the pipeline.

To guide you through this process, this section collects the main points to consider during development.
These points should help in both the development and review process of the new feature.


## Developing a notebook from scratch

When you are starting a blank notebook from scratch you should first
answer few questions:

* How will the notebook be executed:

  - Is it executed interactively by the user?
  - Should it execute the same as the production notebooks through xfel-calibrate CLI.
  - Is it a headless task or a report is needed at the end to see the results.

* Do you need to run the notebook's processing concurrently?
  
  - Is internal concurrency using multiprocessing sufficient, or processing using host level
    cluster computing via SLURM is essential for higher performance.

If `xfel-calibrate` CLI is essential, you need to follow the guidelines in where and how to
write the variables in the first notebook cell and how to include it as one of the CLI
calibration options to execute.

If a report is needed you should make sure to provide sufficient guidance and textual details
using markdown cells and clear prints within the code. You should also structure the notebook cells
it into appropriate subsections.

If SLURM concurrency is needed, you need to identify the variable that the notebook will be
replicated based on to split the processing.

Once you've clarified the above points, you should create a new notebook,
either in an existing detector folder or into a new folder with the new detector's name.
Give it a suffix `_NBC` to denote that it is enabled for the tool chain.

You should then start writing your code following the guidelines
at [How to start writing calibration notebook](how_to_write_xfel_calibrate_notebook_NBC.md)

## Title and Author Information in `markdown` [First Cell]

Especially for report generation the notebook should have a proper title
author and version. These should be given in a leading `markdown` cell in
the form::

    # My Fancy Calculation

    Author](Jane Doe, Version 0.1

    A description of the notebook.

Information in this format allows automatic parsing of author and version.


## Exposing Parameters to the `xfel-calibrate` Command Line [SECOND CELL]

The European XFEL Offline Calibration toolkit automatically deduces
command line arguments from Jupyter notebooks. It does this with an
extended version of nbparameterise_, originally written by Thomas
Kluyver.

Parameter deduction tries to parse all variables defined in the first
code cell of a notebook. The following variable types are supported:

* Numbers](INT or FLOAT
* Booleans
* Strings
* Lists of any of the above

You should avoid having `import` statements in this cell. Line comments
can be used to define the help text provided by the command line interface,
and to signify if lists can be constructed from ranges and if parameters are
required::

    in_folder = ""  # directory to read data from, required
    out_folder = ""  # directory to output to, required
    metadata_folder = ""  # directory containing calibration metadata file when run by xfel-calibrate
    run = [820, ]  # runs to use, required, range allowed
    sequences = [0, 1, 2, 3, 4]  # sequences files to use, range allowed
    modules = [0]  # modules to work on, required, range allowed

    karabo_id = "MID_DET_AGIPD1M-1"  # Detector karabo_id name
    karabo_da = [""]  # a list of data aggregators names, Default [-1] for selecting all data aggregators

    skip-plots = False  # exit after writing corrected files and metadata

The above are some examples of parameters from AGIPD correction notebook.

- Here, `in_folder` and `out_folder` are set as `required` string values. 

Values for required parameters have to be given when executing from the command line. 
This means that any defaults given in the first cell of the code are ignored 
(they are only used to derive the type of the parameter). 

- `modules` and `sequences` are lists of integers, which from the command line could also be assigned using a range expression, 
e.g. `5-10,12,13,18-21`, which would translate to `5,6,7,8,9,12,13,18,19,20`.

.. note::

    nbparameterise_ can only parse the mentioned subset of variable types. An expression
    that evaluates to such a type will not be recognized](e.g. `a = list(range(3))` will
    not work!

- `karabo_id` is a string value indicating the detector to be processed.
- `karabo_da` is a list of strings to indicate the detector's modules to be processed. As `karabo_da` and `modules` 
  are two different variables pointing to the same physical parameter. In the later notebook cells both parameters are synced 
  before usage.

- `skip-plots` is a boolean for skipping the notebook plots to save time and deliver the report as soon as the data are processed. 
  to set `skip-plots` to False from the command line. `--no-skip-plots` is used.

The code would lead to the following parameters being exposed via the command line::

    % xfel-calibrate AGIPD FF --help
      usage](xfel-calibrate [-h] [--no-cluster-job] [--prepare-only] [--report-to str] [--not-reproducible] [--skip-report] [--skip-env-freeze] [--concurrency-par str]
                            [--constants-from str] [--vector-figs] [--slurm-mem int] [--slurm-name str] [--slurm-scheduling int] [--request-time str]
                            [--slurm-partition str] [--reservation str] --in-folder str --out-folder str [--sequences str [str ...]] [--modules str [str ...]]
                            [--train-ids str [str ...]] --run int [--karabo-id str] [--karabo-da str [str ...]] [--receiver-template str] [--path-template str]
                            [--instrument-source-template str] [--index-source-template str] [--ctrl-source-template str] [--karabo-id-control str]
                            [--slopes-ff-from-files str] [--use-dir-creation-date | --no-use-dir-creation-date] [--cal-db-interface str] [--cal-db-timeout int]
                            [--creation-date-offset str] [--mem-cells int] [--bias-voltage int] [--acq-rate float] [--gain-setting int] [--gain-mode int]
                            [--overwrite | --no-overwrite] [--max-pulses int [int ...]] [--mem-cells-db int] [--integration-time int] [--blc-noise-threshold int]
                            [--cm-dark-fraction float] [--cm-dark-range float [float ...]] [--cm-n-itr int] [--hg-hard-threshold int] [--mg-hard-threshold int]
                            [--noisy-adc-threshold float] [--ff-gain float] [--photon-energy float] [--only-offset | --no-only-offset] [--rel-gain | --no-rel-gain]
                            [--xray-gain | --no-xray-gain] [--blc-noise | --no-blc-noise] [--blc-stripes | --no-blc-stripes] [--blc-hmatch | --no-blc-hmatch]
                            [--match-asics | --no-match-asics] [--adjust-mg-baseline | --no-adjust-mg-baseline] [--zero-nans | --no-zero-nans]
                            [--zero-orange | --no-zero-orange] [--blc-set-min | --no-blc-set-min] [--corr-asic-diag | --no-corr-asic-diag]
                            [--force-hg-if-below | --no-force-hg-if-below] [--force-mg-if-below | --no-force-mg-if-below] [--mask-noisy-adc | --no-mask-noisy-adc]
                            [--common-mode | --no-common-mode] [--melt-snow | --no-melt-snow] [--mask-zero-std | --no-mask-zero-std]
                            [--low-medium-gap | --no-low-medium-gap] [--round-photons | --no-round-photons] [--use-ppu-device str] [--ppu-train-offset int]
                            [--use-litframe-finder str] [--litframe-device-id str] [--energy-threshold int] [--use-super-selection str] [--use-xgm-device str]
                            [--recast-image-data str] [--compress-fields str [str ...]] [--skip-plots | --no-skip-plots] [--cell-id-preview int] [--chunk-size int]
                            [--n-cores-correct int] [--n-cores-files int] [--sequences-per-node int] [--max-nodes int] [--max-tasks-per-worker int]
                            DETECTOR TYPE

      # AGIPD Offline Correction #
      ,
      Author](European XFEL Detector Group, Version](2.0,
      ,
      Offline Calibration for the AGIPD Detector

      positional arguments:
        DETECTOR              The detector to calibrate](AGIPD, LPD, PNCCD, GENERIC, TUTORIAL, FASTCCD, JUNGFRAU, GOTTHARD2, EPIX100, EPIX10K, DSSC, REMI, TEST, TEST-
                              RAISES-ERRORS
        TYPE                  Type of calibration.

      optional arguments:
        -h, --help            show this help message and exit
        --sequences str [str ...]
                              sequences to correct, set to -1 for all, range allowed. Default]([-1]
        --modules str [str ...]
                              modules to correct, set to -1 for all, range allowed. Default]([-1]
        --train-ids str [str ...]
                              train IDs to correct, set to -1 for all, range allowed. Default]([-1]
        --karabo-id str       karabo karabo_id. Default](MID_DET_AGIPD1M-1
        --karabo-da str [str ...]
                              a list of data aggregators names, Default [-1] for selecting all data aggregators. Default](['-1']
        --karabo-id-control str
                              karabo-id for control device. Default](MID_EXP_AGIPD1M1
      ...

      required arguments:
        --in-folder str       the folder to read data from, required. Default](None
        --out-folder str      the folder to output to, required. Default](None
        --run int             runs to process, required. Default](None


The following table contains a list of suggested names for certain parameters, allowing
to stay consistent amongst all notebooks.


| Parameter name    | To be used for                                                        | Special purpose            |
| -----------       | --------------------------------------------------------------------- |----------------------------|
| `in_folder`       | the input path data resides in, usually without a run number.         ||
| `out_folder`      | path to write data out to, usually without a run number.              | reports can be placed here |
| `metadata_folder` | directory path for calibration metadata file with local constants.    ||
| `run(s)`          | which XFEL DAQ runs to use, often ranges are allowed.                 ||
| `karabo_id`       | detector karabo name to access detector files and constants.          ||
| `karabo_da`       | refers to detector's modules data aggregator names to process.        ||
| `modules`         | refers to the detector's modules indices to process, ranges often ok. ||
| `sequences`       | sequence files for the XFEL DAQ system, ranges are often ok.          ||
| `local_output`    | write calibration constant from file, not database.                   ||
| `db_output`       | write calibration constant from database, not file.                   |saves the database from unintentional constant |
| `injections`      | developments or testing.                                              |                            |



## Best Coding Practices

In principle there a not restrictions other than that parameters that are exposed to the
command line need to be defined in the first code cell of the notebook.

However, a few guidelines should be observed to make notebook useful for display as
reports and usage by others.

### External Libraries

You may use a wide variety of libraries available in Python, but keep in mind that others
wanting to run the tool will need to install these requirements as well. Thus,

* Do not use a specialized tool if an accepted alternative exists. Plots e.g. should usually
  be created using matplotlib_ and numerical processing should be done in numpy_.

* Keep runtime and library requirements in mind. A library doing its own parallelism either
  needs to programmatically be able to set this up, or automatically do so. If you need to
  start something from the command line first, things might be tricky as you will likely
  need to run this via `POpen` commands with appropriate environment variable.

* Reading out RAW data should be done using extra_data_. It helps in accessing the HDF5 data
  structures efficiently. It reduces the complexity of accessing the RAW or CORRECTED datasets,
  and it provides different methods to select and filter the trains, cells, or pixels of interest.

### Writing out data


If your notebook produces output data, consider writing data out as early as possible,
such that it is available as soon as possible. Detailed plotting and inspection can
be done later on in the notebook.

Also use HDF5 via h5py_ as your output format. If you correct or calibrate
input data, which adheres to the XFEL naming convention, you should maintain the convention
in your output data. You should not touch any data that you do not actively work on and
should assure that the `INDEX` and identifier entries are synchronized with respect to
your output data. E.g. if you remove pulses from a train, the `INDEX/.../count` section
should reflect this.


### Plotting

When creating plots, make sure that the plot is either self-explanatory or add markdown
comments with adequate description. Do not add "free-floating" plots, always put them into
a context. Make sure to label your axes.

Also make sure the plots are readable on an A4-sized PDF page; this is the format the notebook
will be rendered to for report outputs. Specifically, this means that figure sizes should not
exceed approx 15x15 inches.

The report will contain 150 dpi png images of your plots. If you need higher quality output
of individual plot files you should save these separately, e.g. via `fig.savefig(...)` yourself.

## Running a calibration

It utilizes tools such as
[nbconvert](https://github.com/jupyter/nbconvert) and
[nbparameterise](https://github.com/takluyver/nbparameterise) to expose
[Jupyter](http://jupyter.org/) notebooks to a command line interface. In
the process reports are generated from these notebooks.

The general interface is:

    % xfel-calibrate DETECTOR TYPE

where [DETECTOR] and [TYPE] specify the task to
be performed.

Additionally, it leverages the DESY/XFEL Maxwell cluster to run these
jobs in parallel via [SLURM](https://slurm.schedmd.com).

Here is a list of `available_notebooks`.
See the `advanced_topics` if you are
looking for details on how to use as detector group staff.

If you would like to integrate additional notebooks please see the
`development_workflow`.

## Interaction with Calibration Database

During development, it is advised to work with local constant files first before injecting
any calibration constants to the production database. After the notebook's algorithms arguments
matured one can switch over to the test database and then production database.
The reason for this is to avoid injecting wrong constants that can affect production calibration.
And to avoid unnecessary intervention to disable wrong or unused injected calibration constants.

Additionally, the calibration database is limited to XFEL networks, so independent development improves
the workflow.


## Testing

The most important test is that your notebook completes flawlessly outside any special
tool chain feature. After all, the tool chain will only replace parameters, and then
launch a concurrent job and generate a report out of notebook. If it fails to run in the
normal Jupyter notebook environment, it will certainly fail in the tool chain environment.

Once you are satisfied with your current state of initial development, you can add it
to the list of notebooks as mentioned in the :ref:`configuration` section.

Any changes you now make in the notebook will be automatically propagated to the command line.
Specifically, you should verify that all arguments are parsed correctly, e.g. by calling::

    xfel-calibrate DETECTOR NOTEBOOK_TYPE --help

From then on, check include if parallel slurm jobs are executed correctly and if a report
is generated at the end.

Finally, you should verify that the report contains the information you'd like to convey and
is intelligible to people other than you.

???+ note

    You can run the `xfel-calibrate` command without starting a SLURM cluster job, giving
    you direct access to console output, by adding the `--no-cluster-job` option.

## Documenting

Most documentation should be done in the notebook itself. Any notebooks specified in the
`notebook.py` file will automatically show up in the [Available Notebooks](available_notebooks.md) section of this
documentation.


- [nbparameterise](https://github.com/takluyver/nbparameterise)
- [ipcluster](https://ipyparallel.readthedocs.io/en/latest/)
- [matplotlib](https://matplotlib.org/)
- [numpy](http://www.numpy.org/)
- [h5py](https://www.h5py.org/)
- [iCalibrationDB](https://git.xfel.eu/detectors/cal_db_interactive)
- [extra_data](https://extra-data.readthedocs.io/en/latest/)
- [extra-geom](https://extra-geom.readthedocs.io/en/latest/)
- [pasha](https://github.com/European-XFEL/pasha)